﻿using System;
using System.CodeDom.Compiler;
using System.IO;
using System.Reflection;
using System.Text;
using System.Windows.Forms;
using Microsoft.CSharp;
using System.Collections.Generic;

namespace CSharpScript
{
    public class CSharpScriptHost
    {
        /// <summary>
        /// Создает объект CSharpScriptHost
        /// </summary>
        public CSharpScriptHost() {
            Parameters = new Dictionary<string, object>();
        }

        public Dictionary<string, object> Parameters;

        /// <summary>
        /// Показывать окно ожидания в процессе сборки?
        /// </summary>
        public bool ShowWindow { get; set; }

        /// <summary>
        /// Выполняет файл, указанный в filename
        /// </summary>
        /// <param name="filename">Имя файла</param>
        /// <returns></returns>
        public bool ExecuteScriptFromFile(string filename)
        {
            return ExecuteScript(GetFile(filename));
        }
        /// <summary>
        /// Загружает текстовый файл в строковую переменную
        /// </summary>
        /// <param name="filename"></param>
        /// <returns></returns>
        private string GetFile( string filename) {
            string code;
            try {
                StreamReader reader = new StreamReader( filename );
                code = reader.ReadToEnd();
                reader.Close();
                reader.Dispose();
                reader = null;
            }
            catch ( FileNotFoundException fnf ) {
                throw new CSharpScriptHostException( fnf.Message );
            }
            return code;
        }
        private List<string> references;


        /// <summary>
        /// Обработка кода до сборки, формирование списка требующихся дополнительных сборок и выполнение 
        /// команд препроцессора
        /// </summary>
        /// <param name="code">Код скрипта</param>
        /// <returns></returns>
        private string ParseCode(string code) {
            StringBuilder resCode = new StringBuilder();
            foreach ( string line in code.Split( '\n') ) {
                string cmd = line.Trim();
                string[] asm = line.Replace( ";", String.Empty ).Trim().Split( new char[] { ' ', '\t' }, 2 );
                if ( cmd.StartsWith( "#reference" ) ) {
                    if ( asm.Length > 1 )
                        references.Add( asm[1].Trim() );
                    else
                        throw new CSharpScriptHostException( "#reference: Ожидается имя сборки." );
                    continue;
                }
                 
                if ( cmd.StartsWith( "#include" ) ) {
                    if(asm.Length > 1){
                        string file = GetFile( asm[1] );
                        string inccode = ParseCode( file );
                        resultCode.Add( inccode);
                    }
                    else
                        throw new CSharpScriptHostException( "#include: Ожидается имя файла." );
                    continue;
                }
                if ( cmd.StartsWith( "#inline" ) ) {
                    if ( asm.Length > 1 ) {
                        string file = GetFile( asm[1] );
                        resCode.Append(ParseCode( file));
                    }
                    else
                        throw new CSharpScriptHostException( "#inline: Ожидается имя файла." );
                    continue;
                }

                resCode.Append(line);
            }
            return resCode.ToString();
        }
        private List<string> resultCode;

        /// <summary>
        /// Выполняет скрипт
        /// </summary>
        /// <param name="code">Код скрипта</param>
        /// <returns></returns>
        public bool ExecuteScript(string code)
        {
            references = new List<string>();
            resultCode = new List<string>();
            
            FormWait frmWait = new FormWait();
            if ( ShowWindow ) {
                frmWait.Show();
                frmWait.Update();
            }

            resultCode.Add(ParseCode( code ));
            
            CSharpCodeProvider provider = new CSharpCodeProvider();
            string filename = Path.GetTempFileName();
            CompilerParameters parameters = new CompilerParameters();
            parameters.GenerateInMemory = true;
            parameters.GenerateExecutable = false;

            parameters.OutputAssembly = filename;
            foreach ( string asm in references ) {
                parameters.ReferencedAssemblies.Add( asm );
            }
            foreach (Assembly asm in AppDomain.CurrentDomain.GetAssemblies())
                if(parameters.ReferencedAssemblies.IndexOf(asm.ManifestModule.Name) <0 )
                    parameters.ReferencedAssemblies.Add(asm.Location);

            CompilerResults result = provider.CompileAssemblyFromSource(parameters, resultCode.ToArray());
            frmWait.Close();
            frmWait.Dispose();
            frmWait = null;
            StringBuilder errorInfo = new StringBuilder();
            if(result.Errors.Count > 0)
                errorInfo.Append("При сборке возникли ошибки:\n");
            foreach (CompilerError error in result.Errors)
            {
                if (error.IsWarning)
                    continue;
                errorInfo.Append(error.ToString());
                errorInfo.Append("\n");
            }
            if (!string.IsNullOrEmpty(errorInfo.ToString()))
            {
                throw new CSharpScriptHostException( errorInfo.ToString());
            }
            else
            {
                if(File.Exists(filename))
                    File.Delete(filename);
                Type t = result.CompiledAssembly.GetType( "ScriptMain.ScriptMain" );
                MethodInfo m = null;
                if ( t != null ) {
                    m = t.GetMethod( "Main" );
                    if ( m != null ) {
                        m.Invoke( null, new object[] {Parameters } );
                    }
                    else
                        throw new CSharpScriptHostException( "Метод static void Main(Dictionary<string,object>) не найден!" );

                }
                else
                    throw new CSharpScriptHostException( "Не объявлен класс ScriptMain.ScriptMain!" );
                }
               
                return true;
            }
    }
}
